package net.sf.mgp.unclasses;

/*
 * Copyright 2008 MGP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.shared.dependency.analyzer.asm.DependencyClassFileVisitor;
import org.codehaus.plexus.util.DirectoryScanner;

/**
 * Unused classes goal
 * 
 * @goal unclasses
 * @execute phase="compile"
 * 
 * @author Maxim Gordienko
 * 
 */
public class UnclassesMojo extends AbstractMojo {
	private static final int FILE_SUFFIX_LENGTH = ".class".length();

	private static final String[] CLASS_FILE_PATTERNS = { "**/*.class" };

	/**
	 * The input directory for the project classes.
	 * 
	 * @parameter expression="${project.build.outputDirectory}"
	 * @required
	 * @readonly
	 */
	private File classesDirectory;

	public void execute() throws MojoExecutionException {
		try {
			String[] paths = findClassFiles();
			Map<String, String> fileClasses = getFileClassNames(paths);

			Collection<String> unreferenced = getUnreferencedClassNames(fileClasses);
			if (!unreferenced.isEmpty()) {
				logUnreferenced(unreferenced);
			}
		} catch (IOException e) {
			throw new MojoExecutionException("Execution failed", e);
		}
	}

	Collection<String> getUnreferencedClassNames(Map<String, String> fileClassNames) throws IOException {
		DependencyGraph graph = new DependencyGraph(fileClassNames.values());
		for (Entry<String, String> entry : fileClassNames.entrySet()) {
			String classFile = entry.getKey();
			String className = entry.getValue();
			for (String dependency : getDependencies(classFile, className)) {
				if (!className.equals(dependency)) {
					graph.reference(dependency);
				}
			}
		}
		Collection<String> unreferenced = graph.getUnreferenced();
		return unreferenced;
	}

	private String[] findClassFiles() {
		DirectoryScanner scanner = new DirectoryScanner();
		scanner.setBasedir(classesDirectory);
		scanner.setIncludes(CLASS_FILE_PATTERNS);
		scanner.scan();
		return scanner.getIncludedFiles();
	}

	Collection<String> getDependencies(String classFile, String className) throws IOException {
		DependencyClassFileVisitor visitor = new DependencyClassFileVisitor();

		InputStream in = new FileInputStream(new File(classesDirectory, classFile));
		try {
			visitor.visitClass(className, in);
		} finally {
			in.close();
		}
		return visitor.getDependencies();
	}

	private void logUnreferenced(Collection<String> unreferenced) {
		getLog().error("Unreferenced project classes:");
		for (String className : unreferenced) {
			getLog().error(" " + className);
		}
	}

	private Map<String, String> getFileClassNames(String[] paths) {
		Map<String, String> result = new HashMap<String, String>(paths.length);
		for (String path : paths) {
			String className = getClassName(path);
			result.put(path, className);
		}
		return result;
	}

	private String getClassName(String classFilePath) {
		String className = classFilePath.substring(0, classFilePath.length() - FILE_SUFFIX_LENGTH);
		return className.replace('/', '.');
	}
}
